//
// Copyright 2025 Formata, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

pub mod format;
pub use format::*;

pub mod profile;
pub use profile::*;

pub mod docs;
pub use docs::*;

pub mod json;
pub use json::*;

pub mod toml;
pub use toml::*;

pub mod stof;
pub use stof::*;

pub mod yaml;
pub use yaml::*;

pub mod text;
pub use text::*;

pub mod bytes;
pub use bytes::*;

pub mod urlencoded;
pub use urlencoded::*;

#[cfg(feature = "pkg")]
pub mod pkg;
#[cfg(feature = "pkg")]
pub use pkg::*;

#[cfg(feature = "pdf")]
pub mod pdf;
#[cfg(feature = "pdf")]
pub use pdf::*;

#[cfg(feature = "image")]
pub mod image;
#[cfg(feature = "image")]
pub use image::*;

#[cfg(feature = "docx")]
pub mod docx;
#[cfg(feature = "docx")]
pub use docx::*;
