/* tslint:disable */
/* eslint-disable */
export const memory: WebAssembly.Memory;
export const __wbg_stoffunc_free: (a: number, b: number) => void;
export const stoffunc_new: (a: number, b: number, c: number, d: number, e: any, f: number) => number;
export const __wbg_stof_free: (a: number, b: number) => void;
export const start: () => void;
export const stof_binaryExport: (a: number, b: number, c: number, d: any) => [number, number, number];
export const stof_binaryImport: (a: number, b: any, c: number, d: number, e: any, f: number, g: number) => [number, number, number];
export const stof_call: (a: number, b: number, c: number, d: any) => any;
export const stof_get: (a: number, b: number, c: number, d: any) => any;
export const stof_js_library_function: (a: number, b: number) => void;
export const stof_new: () => number;
export const stof_objImport: (a: number, b: any, c: any) => [number, number, number];
export const stof_parse: (a: number, b: number, c: number, d: any, e: number, f: number) => [number, number, number];
export const stof_run: (a: number, b: any) => any;
export const stof_set: (a: number, b: number, c: number, d: any, e: any) => number;
export const stof_stringExport: (a: number, b: number, c: number, d: any) => [number, number, number, number];
export const stof_stringImport: (a: number, b: number, c: number, d: number, e: number, f: any, g: number, h: number) => [number, number, number];
export const stof_sync_call: (a: number, b: number, c: number, d: any) => [number, number, number];
export const stof_sync_run: (a: number, b: any) => [number, number, number, number];
export const wasm_bindgen__convert__closures_____invoke__h53d5cf04cab8438f: (a: number, b: number, c: any) => void;
export const wasm_bindgen__closure__destroy__h72b14ab7db8750ca: (a: number, b: number) => void;
export const wasm_bindgen__convert__closures_____invoke__ha84735728bfe97a9: (a: number, b: number, c: any, d: any) => void;
export const __wbindgen_malloc: (a: number, b: number) => number;
export const __wbindgen_realloc: (a: number, b: number, c: number, d: number) => number;
export const __wbindgen_exn_store: (a: number) => void;
export const __externref_table_alloc: () => number;
export const __wbindgen_externrefs: WebAssembly.Table;
export const __wbindgen_free: (a: number, b: number, c: number) => void;
export const __externref_table_dealloc: (a: number) => void;
export const __wbindgen_start: () => void;
